import uuid
import six.moves.xmlrpc_client as xmlrpclib

from cornice.resource import resource
from cornice.util import extract_json_data
from pyramid.httpexceptions import (
    HTTPCreated,
    HTTPNotFound,
    )
from pyramid_rpc.xmlrpc import xmlrpc_method
import requests

from .models import (
    DBSession,
    Repo,
    User,
    )

from sqlalchemy.orm.exc import NoResultFound


@resource(collection_path='/repos', path='/repos/{name}')
class RepoAPI(object):

    def __init__(self, request):
        self.request = request

    def collection_get(self):
        return {'repos': [repo.name for repo in DBSession.query(Repo)]}

    def collection_post(self):
        name = extract_json_data(self.request).get('name')
        if not name:
            self.request.errors.add('body', 'name', 'name is missing')
            return
        try:
            DBSession.query(Repo).filter(Repo.name == name).one()
        except NoResultFound:
            pass
        else:
            self.request.errors.add('body', 'name', 'name already exists')
            return
        owner_name = extract_json_data(self.request).get('owner')
        if not owner_name:
            self.request.errors.add('body', 'owner', 'owner_name is missing')
            return
        try:
            owner = DBSession.query(User).filter(User.name == owner_name).one()
        except NoResultFound:
            self.request.errors.add('body', 'owner', 'owner does not exist')
            return
        repo = Repo(name=name, owner=owner, turnip_id=str(uuid.uuid4()))
        DBSession.add(repo)
        create_resp = requests.post(
            self.request.registry.settings['turnip.endpoint'] + 'create',
            data={'path': repo.turnip_id})
        if create_resp.status_code != 200:
            raise Exception("turnip failed to create the repository")
        return HTTPCreated(
            self.request.route_url('repoapi', name=repo.name))

    def get(self):
        name = self.request.matchdict['name']
        try:
            DBSession.query(Repo).filter(Repo.name == name).one()
        except NoResultFound:
            return HTTPNotFound('Repository not found')
        return {'name': self.request.matchdict['name']}


@resource(collection_path='/users', path='/users/{name}')
class UserAPI(object):

    def __init__(self, request):
        self.request = request

    def collection_get(self):
        return {'users': [user.name for user in DBSession.query(User)]}

    def collection_post(self):
        name = extract_json_data(self.request).get('name')
        if not name:
            self.request.errors.add('body', 'name', 'name is missing')
            return
        password = extract_json_data(self.request).get('password')
        if not name:
            self.request.errors.add('body', 'password', 'password is missing')
            return
        try:
            DBSession.query(User).filter(User.name == name).one()
        except NoResultFound:
            pass
        else:
            self.request.errors.add('body', 'name', 'name already exists')
            return
        user = User(name=name, password=password)
        DBSession.add(user)
        return HTTPCreated(
            self.request.route_url('userapi', name=user.name))

    def get(self):
        name = self.request.matchdict['name']
        try:
            DBSession.query(User).filter(User.name == name).one()
        except NoResultFound:
            return HTTPNotFound('User not found')
        return {'name': self.request.matchdict['name']}


@xmlrpc_method(endpoint='githosting')
def translatePath(request, path, permission, authenticated_uid,
                  can_authenticate):
    try:
        repo = DBSession.query(Repo).filter(
            Repo.name == path.lstrip('/')).one()
    except NoResultFound:
        raise xmlrpclib.Fault(1, "Repo does not exist")

    writable = authenticated_uid == repo.owner_id

    if permission != b'read' and not writable:
        if not can_authenticate or authenticated_uid is not None:
            raise xmlrpclib.Fault(2, "Repo is read-only")
        else:
            raise xmlrpclib.Fault(3, "Authorisation required")

    return {'path': repo.turnip_id, 'writable': writable}


@xmlrpc_method(endpoint='githosting')
def authenticateWithPassword(request, username, password):
    try:
        user = DBSession.query(User).filter(User.name == username).one()
    except NoResultFound:
        raise xmlrpclib.Fault(3, "Invalid username or password")
    if password != user.password:
        raise xmlrpclib.Fault(3, "Invalid username or password")
    return {'user': username, 'uid': user.id}
